/*
 *----------------------------------------------------------------------
 *
 * File: Callback.c
 *
 * Description: This file contains the callback facility processing.  
 * 
 * Exported functions:
 *   soar_add_callback
 *   soar_invoke_callbacks
 *   soar_remove_callback
 *
 *   soar_callback_data_free_string
 *   soar_callback_enum_to_name
 *   soar_callback_name_to_enum
 *
 * Each agent has a separate callback table.  The table has one entry
 * per callback type and the entry is a pointer to a list.  The list
 * contains installed callbacks, one callback per list cons cell.
 *
 *----------------------------------------------------------------------
 */

#include "soarkernel.h"
#include "callback.h"

char * soar_callback_names[] = {    /* Must match order of       */
  "none",                           /* SOAR_CALLBACK_TYPE        */
  "system-startup",
  "system-termination",
  "after-init-agent",
  "before-init-soar",
  "after-init-soar",
  "after-halt-soar",
  "before-schedule-cycle",
  "after-schedule-cycle",
  "before-decision-cycle",
  "after-decision-cycle",
  "before-input-phase",
  "input-phase-cycle",
  "after-input-phase",
  "before-preference-phase-cycle",
  "after-preference-phase-cycle",
  "before-wm-phase-cycle",
  "after-wm-phase-cycle",
  "before-output-phase",
  "output-phase",
  "after-output-phase",
  "before-decision-phase-cycle",
  "after-decision-phase-cycle",
  "wm-changes",
  "create-new-context",
  "pop-context-stack",
  "create-new-attribute-impasse",
  "remove-attribute-impasse",
  "production-just-added",
  "production-just-about-to-be-excised",
  "firing",
  "retraction",
  "system-parameter-changed",
  "print",
  "log",
  /* Nothing corresponds to NUMBER_OF_CALLBACKS */
};


void soar_init_callbacks (soar_callback_agent the_agent)
{
  SOAR_CALLBACK_TYPE ct;

  for (ct = 1; ct < NUMBER_OF_CALLBACKS; ct++)
    {
      ((agent *)the_agent)->soar_callbacks[ct] = (list *) NIL;
    }
}

void soar_add_callback (soar_callback_agent the_agent, 
			SOAR_CALLBACK_TYPE callback_type, 
			soar_callback_fn fn, 
			soar_callback_data data,
			soar_callback_free_fn free_fn,
			soar_callback_id id)
{
  soar_callback * cb;

  cb = (soar_callback *) malloc (sizeof(soar_callback));
  cb->function      = fn;
  cb->data          = data;
  cb->free_function = free_fn;
  cb->id            = savestring((char *)id);
  
  push(cb, ((agent *) the_agent)->soar_callbacks[callback_type]);
}

void soar_callback_data_free_string (soar_callback_data data)
{
  free((char *) data);
}

char * soar_callback_enum_to_name (SOAR_CALLBACK_TYPE i, 
				   bool monitorable_only)
{
  int limit;

  if (monitorable_only)
    {
      limit = NUMBER_OF_MONITORABLE_CALLBACKS;
    }
  else 
    {
      limit = NUMBER_OF_CALLBACKS;
    }

  if ((0 < i) && (i < limit))
    {
      return soar_callback_names[i];
    }
  return NULL;
}

SOAR_CALLBACK_TYPE soar_callback_name_to_enum (char * name,
					       bool monitorable_only)
{
  int limit;
  SOAR_CALLBACK_TYPE i;

  if (monitorable_only)
    {
      limit = NUMBER_OF_MONITORABLE_CALLBACKS;
    }
  else 
    {
      limit = NUMBER_OF_CALLBACKS;
    }

  for(i = 1; i < limit; i++)
    {
      if (!strcmp(name, soar_callback_names[i]))
	{
	  return i;
	}
    }

  return NO_CALLBACK;
}

bool soar_exists_callback(soar_callback_agent the_agent,
			  SOAR_CALLBACK_TYPE callback_type)
{
  list * cb_cons;

  cb_cons = ((agent *)the_agent)->soar_callbacks[callback_type];

  if (cb_cons == NULL)
    {
      return FALSE;
    }

  return TRUE;
}

soar_callback * soar_exists_callback_id (soar_callback_agent the_agent,
					 SOAR_CALLBACK_TYPE callback_type,
					 soar_callback_id id)
{
  cons * c;

  for (c = ((agent *)the_agent)->soar_callbacks[callback_type];
       c != NIL;
       c = c->rest)
    {
      soar_callback * cb;

      cb = (soar_callback *) c->first;

      if (!strcmp(cb->id, id))
	{
	  return cb;
	}
    }

  return NULL;
}

void soar_destroy_callback(soar_callback * cb)
{
  if (cb->id)
    {
      free(cb->id);
    }
  if (cb->free_function)
    {
      cb->free_function(cb->data);
    }
  free((void *) cb);
}

void soar_invoke_callbacks (soar_callback_agent the_agent, 
			    SOAR_CALLBACK_TYPE callback_type, 
			    soar_call_data call_data)
{
  cons * c;

  for (c = ((agent *)the_agent)->soar_callbacks[callback_type];
       c != NIL;
       c = c->rest)
    {
      soar_callback * cb;

      cb = (soar_callback *) c->first;
      cb->function(the_agent, cb->data, call_data);
    }
}

void soar_invoke_first_callback (soar_callback_agent the_agent, 
				 SOAR_CALLBACK_TYPE callback_type, 
				 soar_call_data call_data)
{
  list * head;
 
  head = ((agent *)the_agent)->soar_callbacks[callback_type];

  if (head != NULL)
    {
      soar_callback * cb;

      cb = (soar_callback *) head->first;
      cb->function(the_agent, cb->data, call_data);
    }
}

void soar_list_all_callbacks (soar_callback_agent the_agent,
			      bool monitorable_only)
{
  int limit;
  SOAR_CALLBACK_TYPE ct;

  if (monitorable_only)
    {
      limit = NUMBER_OF_MONITORABLE_CALLBACKS;
    }
  else 
    {
      limit = NUMBER_OF_CALLBACKS;
    }

  for (ct = 1; ct < limit; ct++)
    {
      print("%s: ", soar_callback_enum_to_name(ct, FALSE));
      soar_list_all_callbacks_for_event (the_agent, ct);
      print("\n");
    }
}

void soar_list_all_callbacks_for_event (soar_callback_agent the_agent,
					SOAR_CALLBACK_TYPE ct)
{
  cons * c;

  for (c = ((agent *)the_agent)->soar_callbacks[ct]; 
       c != NIL; 
       c = c->rest)
    {
      soar_callback * cb;
      
      cb = (soar_callback *) c->first;

      print("%s ", cb->id);
    }
}

void soar_pop_callback (soar_callback_agent the_agent, 
			SOAR_CALLBACK_TYPE callback_type)
{
  list * head;
  soar_callback * cb;

  head = ((agent *)the_agent)->soar_callbacks[callback_type];
  
  if (head == NULL)
    {
      print_string("Attempt to remove non-existant callback.\n");
      return;
    }

  if (   (callback_type == PRINT_CALLBACK)
      && (head->rest == NULL))
    {
      print_string("Attempt to remove last print callback. Ignored.\n");
      return;
    }

  cb = (soar_callback *) head->first;

  ((agent *)the_agent)->soar_callbacks[callback_type] = head->rest;
  soar_destroy_callback(cb);
  free_cons(head);
}

void soar_push_callback (soar_callback_agent the_agent, 
			SOAR_CALLBACK_TYPE callback_type, 
			soar_callback_fn fn, 
			soar_callback_data data,
			soar_callback_free_fn free_fn)
{
  soar_callback * cb;

  cb = (soar_callback *) malloc (sizeof(soar_callback));
  cb->function      = fn;
  cb->data          = data;
  cb->free_function = free_fn;
  cb->id            = NULL;
  
  push(cb, ((agent *) the_agent)->soar_callbacks[callback_type]);
}

void soar_remove_all_monitorable_callbacks (soar_callback_agent the_agent)
{
  SOAR_CALLBACK_TYPE ct;

  for (ct = 1; ct < NUMBER_OF_MONITORABLE_CALLBACKS; ct++)
    {
      soar_remove_all_callbacks_for_event (the_agent, ct);
    }
}

void soar_remove_all_callbacks_for_event (soar_callback_agent the_agent,
					  SOAR_CALLBACK_TYPE ct)
{
  cons * c;
  list * next;

  next = ((agent *)the_agent)->soar_callbacks[ct];

  for (c = next; c != NIL; c = next)
    {
      soar_callback * cb;
      
      cb = (soar_callback *) c->first;
	  
      next = next->rest;
      soar_destroy_callback(cb);
      free_cons(c);
    }

  ((agent *)the_agent)->soar_callbacks[ct] = NIL;
}

void soar_remove_callback (soar_callback_agent the_agent, 
			   SOAR_CALLBACK_TYPE callback_type, 
			   soar_callback_id id)
{
  cons * c;
  cons * prev_c = NULL;     /* Initialized to placate gcc -Wall */
  list * head;

  head = ((agent *)the_agent)->soar_callbacks[callback_type];

  for (c = head; c != NIL; c = c->rest)
    {
      soar_callback * cb;

      cb = (soar_callback *) c->first;

      if (!strcmp(cb->id, id))
	{
	  if (c != head)
	    {
	      prev_c->rest = c->rest;
	      soar_destroy_callback(cb);
	      free_cons(c);
	      return;
	    }
	  else
	    {
	      ((agent *)the_agent)->soar_callbacks[callback_type] 
		= head->rest;
	      soar_destroy_callback(cb);
	      free_cons(c);
	      return;
	    }
	}
      prev_c = c;
    }
}

void soar_callback_test_callback (soar_callback_agent the_agent,
				  soar_callback_data data,
				  soar_call_data call_data)
{
  printf("%s test callback executed.\n", (char *) data);
}


void soar_test_all_monitorable_callbacks(soar_callback_agent the_agent)
{
  SOAR_CALLBACK_TYPE i;
  static char * test_callback_name = "test";

  for(i = 1; i < NUMBER_OF_MONITORABLE_CALLBACKS; i++)
    {
      soar_add_callback(the_agent, i, 
			(soar_callback_fn) soar_callback_test_callback,
			soar_callback_enum_to_name(i, TRUE), 
			NULL, test_callback_name);
    }
}
