#include <soar.h>
#include <soarScheduler.h>
#include <soarCommandUtils.h>

/* Globals vars (with respect to this file) */

RegisteredInterp *registry = NULL;  /* List of all interpreters
				     * registered by this process. */

/* Forward decls of prototypes */

#ifdef USE_TK
int Soar_AddTkExtension _ANSI_ARGS_((char * name, Tcl_Interp * new_interp, 
				     option_table * options, 
				     char * process_name));
#endif

int Soar_AddSoarExtension _ANSI_ARGS_((char * name, Tcl_Interp * new_interp, 
				      option_table * options, 
				      agent ** return_agent));
int CreateInterpCmd _ANSI_ARGS_((ClientData clientData, Tcl_Interp * interp,
				 int argc, char *argv[]));
int DestroyInterpCmd _ANSI_ARGS_((ClientData clientData, Tcl_Interp * interp,
				  int argc, char *argv[]));
int ListInterpsCmd _ANSI_ARGS_((ClientData clientData, Tcl_Interp * interp,
				int argc, char *argv[]));
int ScheduleInterpCmd _ANSI_ARGS_((ClientData clientData, Tcl_Interp * interp,
				   int argc, char *argv[]));
int SelectInterpCmd _ANSI_ARGS_((ClientData clientData, Tcl_Interp * interp,
				 int argc, char *argv[]));
int SoarnewsCmd _ANSI_ARGS_((ClientData clientData, Tcl_Interp * interp,
			     int argc, char *argv[]));
int VersionCmd _ANSI_ARGS_((ClientData clientData, Tcl_Interp * interp,
			    int argc, char *argv[]));

/* External decls of prototypes */
#ifdef USE_TK
int Tk_SendCmd _ANSI_ARGS_((ClientData clientData, Tcl_Interp * interp, 
			    int argc, char * argv[]));
#else
int Tcl_SendCmd _ANSI_ARGS_((ClientData clientData, Tcl_Interp * interp, 
			     int argc, char * argv[]));
#endif

/*
 *----------------------------------------------------------------------
 *
 * Soar_CreateInterpreter --
 *
 *	This procedure creates a new Tcl interpreter and adds the
 *      implied extensions.
 *
 * Results:
 *	None.
 *
 * Side effects:
 *	A new interpreter is created and registered for sends.
 *
 *----------------------------------------------------------------------
 */
int 
Soar_CreateInterpreter(name, interp_type, options, process_name, tty,
		       return_interp)
     char * name;
     int interp_type;
     option_table * options;
     char * process_name;
     int tty;
     Tcl_Interp ** return_interp;
{
  agent * new_agent = NULL;
  agent * prev_agent = soar_agent;
  Tcl_Interp * new_interp;

  new_interp = Tcl_CreateInterp();
  *return_interp = new_interp;

#ifdef TCL_MEM_DEBUG
  Tcl_InitMemory(new_interp);
#endif

  /*
   * Set the "tcl_interactive" variable.
   */

  Tcl_SetVar(new_interp, "tcl_interactive",
	     ((options->fileName == NULL) && tty) ? "1" : "0", 
	     TCL_GLOBAL_ONLY);

  Tcl_SetVar(new_interp, "interp_type", "tclsh", TCL_GLOBAL_ONLY);

  Tcl_SetVar(new_interp, "tcl_man_dir", TCL_MAN_DIR, TCL_GLOBAL_ONLY);

  if (Tcl_Init(new_interp) != TCL_OK) {
    fprintf(stderr, "Tcl_Init failed for %s: %s\n", name, new_interp->result);
    return TCL_ERROR;
  }

  Tcl_CreateCommand(new_interp, "create-interp", CreateInterpCmd, NULL, NULL);
  Tcl_CreateCommand(new_interp, "destroy-interp", DestroyInterpCmd, 
		    NULL, NULL);
  Tcl_CreateCommand(new_interp, "list-interps", ListInterpsCmd, NULL, NULL);
  Tcl_CreateCommand(new_interp, "schedule-interp", ScheduleInterpCmd, 
		    NULL, NULL);
  Tcl_CreateCommand(new_interp, "select-interp", SelectInterpCmd, NULL, NULL);
  Tcl_CreateCommand(new_interp, "soarnews", SoarnewsCmd, NULL, NULL);
  Tcl_CreateCommand(new_interp, "version", VersionCmd, NULL, NULL);

  if (   options->tk_enabled 
      && (   (interp_type == SOAR_ARGV_WISH)
	  || (interp_type == SOAR_ARGV_AGENT)))
    {
      /* Add Tk Extension.  This also registers the interp so that
	 remote "send"s will work. */
#ifdef USE_TK
      if (Soar_AddTkExtension (name, new_interp, options, process_name) 
	  != TCL_OK)
	{
	  return TCL_ERROR;
	}
      Tcl_SetVar(new_interp, "interp_type", "wish", TCL_GLOBAL_ONLY);

      Tcl_SetVar(new_interp, "tk_man_dir", TK_MAN_DIR, TCL_GLOBAL_ONLY);
      
      /* Since Tk interp registration may have led to a name change,
         update the name to reflect the actual registered one. */

      {
	RegisteredInterp * riPtr;

	riPtr = Soar_GetRegisteredInterpByInterp(new_interp);
	name = riPtr->name;
      }
#endif
    }
  else
    {
      /* For non-Tk shells, we add a register entry by hand so that
	 local sends will work for vanilla Tcl interps too.         */

      RegisteredInterp * riPtr;

      riPtr = Soar_MakeRegisterEntry(name, new_interp, NULL, NULL);
#ifdef USE_TK
      Tcl_CreateCommand(new_interp, "send", Tk_SendCmd, 
			(ClientData) riPtr, NULL);
#else
      Tcl_CreateCommand(new_interp, "send", Tcl_SendCmd, 
			(ClientData) riPtr, NULL);
#endif
    }

  /* Make sure we do this after any possible name changes */

  Tcl_SetVar(new_interp, "interp_name", name, TCL_GLOBAL_ONLY);

  if (interp_type == SOAR_ARGV_AGENT)
    {
      Soar_AddSoarExtension (name, new_interp, options, &new_agent);
      Soar_UpdateRegistryAgent (name, new_agent);
      Tcl_SetVar(new_interp, "interp_type", "agent", TCL_GLOBAL_ONLY);

      /* Some of the following code presumes that soar_agent has been
	 set globally.  We need to do this temporarily in this function
	 to get the agent code setup properly. */

      soar_agent = new_agent;
    }

  { char * libDir;

    libDir = getenv("SOAR_LIBRARY");

    if (libDir == NULL) {
      fprintf(stderr, 
	      "Can't find $soar_library/soar.tcl; perhaps you\n");
      fprintf(stderr, 
	      "need to install Soar or set your SOAR_LIBRARY\n");
      fprintf(stderr, 
	      "environment variable?\n");
      Tcl_SetVar(new_interp, "soar_library", "", TCL_GLOBAL_ONLY);
      Tcl_SetVar(new_interp, "default", "default.soar", TCL_GLOBAL_ONLY);
    } else {
      Tcl_DString path;

      Tcl_SetVar(new_interp, "soar_library", libDir, TCL_GLOBAL_ONLY);

      Tcl_DStringInit(&path);
      Tcl_DStringAppend(&path, libDir, -1);
      Tcl_DStringAppend(&path, "/default.soar", -1);
      Tcl_SetVar(new_interp, "default", Tcl_DStringValue(&path), 
		 TCL_GLOBAL_ONLY);
      Tcl_DStringFree(&path);
    }
  }
  
  if (Soar_Init(new_interp) != TCL_OK) {
    fprintf(stderr, "Soar_Init failed for %s: %s\n", 
	    name, new_interp->result);
    soar_agent = prev_agent;
    return TCL_ERROR;
  }
   
  /* Add any C extensions for the agent */
  Soar_InterpInit(new_interp, interp_type, new_agent);
    
  {
    int code;
    static char cmd[] = "init-interpreter";

    Tcl_SetVar(new_interp, "path", 
	       (options->path == NULL) ? "" : options->path, 
	       TCL_GLOBAL_ONLY);
    
    code = Tcl_GlobalEval(new_interp, cmd);

    if (code != TCL_OK)
      {
	fprintf(stderr, "Error: Failed eval of \"%s\"", cmd);
	fprintf(stderr, "Reason: %s\n", new_interp->result);
      }
  }

  soar_agent = prev_agent;
  return TCL_OK;
}

/*
 *--------------------------------------------------------------
 *
 * Soar_MakeRegisterEntry --
 *
 *	This procedure is called to create a new register
 *      entry.
 *
 * Results:
 *	None.
 *
 * Side effects:
 *	Registration info is saved, thereby allowing the
 *	"send" command to be used later to invoke commands
 *	in the interpreter.  The registration will be removed
 *	automatically when the interpreter is deleted.
 *
 *--------------------------------------------------------------
 */

RegisteredInterp *
Soar_MakeRegisterEntry (name, interp, dispPtr, agentPtr)
    char *name;			/* The name that will be used to
				 * refer to the interpreter in later
				 * "send" commands.  Must be globally
				 * unique. */     
    Tcl_Interp *interp;		/* Interpreter associated with name. */
    ClientData dispPtr;         /* Display for Tk windows */
    agent * agentPtr;    
{
    RegisteredInterp *riPtr;

    /*
     * Add an entry in the local registry of names owned by this
     * process.
     */

    riPtr = (RegisteredInterp *) ckalloc(sizeof(RegisteredInterp));

    if (name != NULL) {
      riPtr->name = (char *) ckalloc((unsigned) (strlen(name) + 1));
      strcpy(riPtr->name, name);
    }

    riPtr->interp   = interp;
    riPtr->dispPtr  = dispPtr;
    riPtr->agentPtr = agentPtr;
    riPtr->nextPtr  = registry;
    registry = riPtr;

    return riPtr;
  }

/*
 *--------------------------------------------------------------
 *
 * Soar_RegisterArgv --
 *
 *	This procedure is called to add argv info to the 
 *      registered interpreters up to the indicated one.
 *
 * Results:
 *	None.
 *
 * Side effects:
 *	Adds argv info to the most recent batch of registered 
 *      interps.
 *
 *--------------------------------------------------------------
 */

void
Soar_RegisterArgv (argc, argv, end)
     int argc;
     char ** argv;
     RegisteredInterp *end;
{
  RegisteredInterp *riPtr;
  char * args;
  char buf[20];

  for (riPtr = registry; riPtr != end; riPtr = riPtr->nextPtr) 
    {
      args = Tcl_Merge(argc-1, argv+1);
      Tcl_SetVar(riPtr->interp, "argv", args, TCL_GLOBAL_ONLY);
      ckfree(args);
      sprintf(buf, "%d", argc-1);
      Tcl_SetVar(riPtr->interp, "argc", buf, TCL_GLOBAL_ONLY);
    }
}

/*
 *--------------------------------------------------------------
 *
 * Soar_GetRegisteredInterp --
 *
 *	This procedure is called to get a registered interp by
 *      name.
 *
 * Results:
 *	NULL if the name is not in the registry, a pointer to
 *      the registered interp if it is.
 *
 * Side effects:
 *	None.
 *
 *--------------------------------------------------------------
 */

RegisteredInterp * 
Soar_GetRegisteredInterp (name)
     char * name;
{
  RegisteredInterp *riPtr;

  for (riPtr = registry; riPtr != NULL; riPtr = riPtr->nextPtr) {
    if (!strcmp(riPtr->name, name)) {
      return riPtr;
    }
  }
  
  return NULL;
}

/*
 *--------------------------------------------------------------
 *
 * Soar_GetRegisteredInterpByInterp --
 *
 *	This procedure is called to get a registered interp by
 *      its Tcl interp struct.
 *
 * Results:
 *	NULL if the name is not in the registry, a pointer to
 *      the registered interp if it is.
 *
 * Side effects:
 *	None.
 *
 *--------------------------------------------------------------
 */

RegisteredInterp * 
Soar_GetRegisteredInterpByInterp (interp)
     Tcl_Interp * interp;
{
  RegisteredInterp *riPtr;

  for (riPtr = registry; riPtr != NULL; riPtr = riPtr->nextPtr) {
    if (interp == riPtr->interp) {
      return riPtr;
    }
  }
  
  return NULL;
}

/*
 *--------------------------------------------------------------
 *
 * Soar_SetRegisteredInterp --
 *
 *	This procedure is called to set the globally active 
 *      registered interp by name.
 *
 * Results:
 *	NULL if the name is not in the registry, a pointer to
 *      the registered interp if it is.
 *
 * Side effects:
 *	None.
 *
 *--------------------------------------------------------------
 */

RegisteredInterp * 
Soar_SetRegisteredInterp (name)
     char * name;
{
  RegisteredInterp *riPtr;

  for (riPtr = registry; riPtr != NULL; riPtr = riPtr->nextPtr) {
    if (!strcmp(riPtr->name, name)) {
      mainInterp = riPtr;
      return riPtr;
    }
  }
  
  return NULL;
}

/*
 *--------------------------------------------------------------
 *
 * Soar_NameInRegistery --
 *
 *	This procedure is called to determine if the named
 *      interpreter is already present in the registry.
 *
 * Results:
 *	0 if the name is not in the registry, non-zero if it is.
 *
 * Side effects:
 *	None.
 *
 *--------------------------------------------------------------
 */

int
Soar_NameInRegistry (name)
     char * name;
{
  RegisteredInterp *riPtr;

  for (riPtr = registry; riPtr != NULL; riPtr = riPtr->nextPtr) {
    if (!strcmp(riPtr->name, name)) {
      return 1;
    }
  }
  
  return 0;
}

/*
 *--------------------------------------------------------------
 *
 * Soar_UpdateRegisteryAgent --
 *
 *	This procedure is called to update an existing registry
 *      entry to include a pointer to a corresponding agent.
 *
 * Results:
 *	0 if the name is not in the registry, non-zero if it is.
 *
 * Side effects:
 *	Updates registered entry with given args.
 *
 *--------------------------------------------------------------
 */

int
Soar_UpdateRegistryAgent (name, agentPtr)
    char *name;			/* The name that will be used to
				 * refer to the interpreter in later
				 * "send" commands.  Must be globally
				 * unique. */     
    agent * agentPtr;     
{
  RegisteredInterp *riPtr;

  for (riPtr = registry; riPtr != NULL; riPtr = riPtr->nextPtr) {
    if (!strcmp(riPtr->name, name)) {
      riPtr->agentPtr = agentPtr;
      return 1;
    }
  }
  
  return 0;
}

/*
 *--------------------------------------------------------------
 *
 * Soar_DestroyRegisterEntry --
 *
 *	This procedure is called to remove a registered interp.
 *
 * Results:
 *	None.
 *
 * Side effects:
 *	Removes the indicated entry, if present.
 *
 *--------------------------------------------------------------
 */

void
Soar_DestroyRegisterEntry (riPtrToRemove)
  RegisteredInterp *riPtrToRemove;
{
  RegisteredInterp *riPtr;
  RegisteredInterp *riPtrPrev = NULL;

  for (riPtr = registry; riPtr != NULL; riPtr = riPtr->nextPtr) {
    if (riPtr == riPtrToRemove) {
      if (riPtr == registry)
	{
	  registry = riPtr->nextPtr;
	}
      else
	{
	  riPtrPrev->nextPtr = riPtr->nextPtr;
	}
      ckfree((char *) riPtr->name);
      ckfree((char *) riPtr);
    }
    riPtrPrev = riPtr;
  }
}

/*
 *--------------------------------------------------------------
 *
 * Soar_AddTkExtension --
 *
 *	This procedure is called to add the Tk extension to a
 *      Tcl interpreter.
 *
 * Results:
 *	A normal Tcl return code.
 *
 * Side effects:
 *	Adds Tk commands to the given Tcl interpreter.
 *
 *--------------------------------------------------------------
 */

#ifdef USE_TK
int
Soar_AddTkExtension (name, new_interp, options, process_name)
     char * name;
     Tcl_Interp * new_interp;
     option_table * options;
     char * process_name;
{
  Tk_Window mainWindow;

  /*
   * If a display was specified, put it into the DISPLAY
   * environment variable so that it will be available for
   * any sub-processes created by us.
   */

  if (options->display != NULL) {
    Tcl_SetVar2(new_interp, "env", "DISPLAY", options->display, 
		TCL_GLOBAL_ONLY);
  }

  /*
   * Initialize the Tk application.
   */
  
  mainWindow = Tk_CreateMainWindow(new_interp, options->display, name, "Tk");
  if (mainWindow == NULL) {
    fprintf(stderr, "%s\n", new_interp->result);
    return TCL_ERROR;
  }
  if (options->synchronize) {
    XSynchronize(Tk_Display(mainWindow), True);
  }
  Tk_GeometryRequest(mainWindow, 200, 200);
  
  Tcl_SetVar(new_interp, "argv0", 
	     (options->fileName != NULL) ? options->fileName : process_name,
	     TCL_GLOBAL_ONLY);
  
  /*
   * Also set the "geometry" variable from the geometry specified on 
   * the command line.
   */
  
  if (options->geometry != NULL) {
    Tcl_SetVar(new_interp, "geometry", options->geometry, TCL_GLOBAL_ONLY);
  }
  
  /*
   * Invoke application-specific initialization.
   */
  
  if (Tk_Init(new_interp) != TCL_OK) {
    fprintf(stderr, "Tk_Init failed for %s: %s\n", name, new_interp->result);
    return TCL_ERROR;
  }
  
  /*
   * Set the geometry of the main window, if requested.
   */
  
  if (options->geometry != NULL) 
    {
      int code;

      code = Tcl_VarEval(new_interp, "wm geometry . ", 
			 options->geometry, (char *) NULL);
      if (code != TCL_OK) {
	fprintf(stderr, "%s\n", new_interp->result);
      }
    }

  return TCL_OK;
}
#endif

/*
 *--------------------------------------------------------------
 *
 * Soar_AddSoarExtension --
 *
 *	This procedure is called to add the Soar extension to a
 *      Tcl interpreter.
 *
 * Results:
 *	A normal Tcl return code.
 *
 * Side effects:
 *	Adds Soar commands to the given Tcl interpreter.
 *
 *--------------------------------------------------------------
 */

int
Soar_AddSoarExtension (name, new_interp, options, return_agent)
     char * name;
     Tcl_Interp * new_interp;
     option_table * options;
     agent ** return_agent;
{
  agent * new_agent;
  
  new_agent = create_soar_agent (name);
  *return_agent = new_agent;

  new_agent->interpreter = new_interp;
  
  if (options->tk_enabled)
    {
      /* 
       * Don't show main window ...
       * The window can be shown again using either:
       *   wm deiconify  - shows open window
       *   wm iconify    - shows icon for window
       */

      {
	static char cmd[] = "wm withdraw .";
	int code;

	code = Tcl_GlobalEval(new_interp, cmd); 
	    
	if (code != TCL_OK)
	  {
	    fprintf(stderr, "Error: Failed eval of \"%s\"", cmd);
	    fprintf(stderr, "Reason: %s\n", new_interp->result);
	  }
      }

    } /* end of tk_enabled */
  
  new_agent->tcl_output_buffer = NIL;
  
  Soar_InstallCommands(new_agent);
  
  return TCL_OK;
}

/*
 *----------------------------------------------------------------------
 *
 * Soar_Init --
 *
 *      This procedure is typically invoked by Tcl_AppInit procedures
 *      to perform additional Soar initialization for a Tcl interpreter,
 *      such as sourcing the "soar.tcl" script.
 *
 * Results:
 *      Returns a standard Tcl completion code and sets interp->result
 *      if there is an error.
 *
 * Side effects:
 *      Depends on what's in the soar.tcl script.
 *
 *----------------------------------------------------------------------
 */
  

int
Soar_Init(interp)
    Tcl_Interp *interp;         /* Interpreter to initialize. */
{
  static char initCmd[] =
    "if [file exists $soar_library/soar.tcl] {\n\
	    source $soar_library/soar.tcl\n\
	} else {\n\
	    set msg \"can't find $soar_library/soar.tcl; perhaps you \"\n\
	    append msg \"need to\\ninstall Soar or set your SOAR_LIBRARY \"\n\
	    append msg \"environment variable?\"\n\
	    error $msg\n\
	}";
 	
    return Tcl_GlobalEval(interp, initCmd);
}

/*
 *----------------------------------------------------------------------
 *
 * SelectInterpCmd --
 *
 *	This is the command procedure for the "select-interp" command, 
 *	which selects a Tcl interpreter (and its corresponding agent,
 *      if any) to be used at the global level.
 *
 * Syntax:  select-interp name
 *
 * Results:
 *	Returns a standard Tcl completion code.
 *
 * Side effects:
 *	Selects a Tcl interpreter (and possibly a Soar agent) to use 
 *      at the global level.
 *
 *----------------------------------------------------------------------
 */

int SelectInterpCmd (ClientData clientData, 
		    Tcl_Interp * interp,
		    int argc, char *argv[])
{
  char * name;
  
  if (argc > 2)
    {
      interp->result = "Too many arguments";
      return TCL_ERROR;
    }

  /* If no name given, then none can be selected. */
  if (argc < 2)
    {
      interp->result = "Expected name but none given.";

      return TCL_ERROR;
    }

  name = argv[1];
  
  /* Check to see if name matches an existing interpreter */

  if (Soar_GetRegisteredInterp(name) == NULL) 
    {
      /* No agent found, the name doesn't match any agents */
      sprintf(interp->result, "Unknown name given: %s.", argv[1]);
      return TCL_ERROR;      
    }

  Soar_SelectGlobalInterp(name);
  return TCL_OK;
}

/*
 *----------------------------------------------------------------------
 *
 * CreateInterpCmd --
 *
 *	This is the command procedure for the "create-interp" command, 
 *	which creates a Tcl interpreter.
 * 
 * Syntax: create-interp type name
 *
 * Results:
 *	Returns a standard Tcl completion code.
 *
 * Side effects:
 *	Creates a new Tcl interpreter with apropriate extensions.
 *
 *----------------------------------------------------------------------
 */

int 
CreateInterpCmd (ClientData clientData, 
		 Tcl_Interp * interp,
		 int argc, char *argv[])
{
  RegisteredInterp * last_registered_interp;

  if (argc < 2)
    {                                    /* no arguments to command    */
      sprintf(interp->result, "Too few arguments");
      return TCL_ERROR;
    }

  /* Note the most recently created interpreter. */
  last_registered_interp = registry;

  if (Soar_DefineInterps(&argc, argv, interp) != TCL_OK)
    {
      return TCL_ERROR;
    }

  /* 
   * Add all unrecognized argv switches to registered interpreters 
   */
    
  Soar_RegisterArgv(argc, argv, last_registered_interp);

  return TCL_OK;
}

/*
 *----------------------------------------------------------------------
 *
 * DestroyInterpCmd --
 *
 *	This is the command procedure for the "destroy-interp" command, 
 *	which destroys a Tcl interpreter
 *
 * Syntax:  destroy-interp name
 *
 * Results:
 *	Returns a standard Tcl completion code.
 *
 * Side effects:
 *	Destroys a Tcl interpreter and its associated agent, if any.
 *
 *----------------------------------------------------------------------
 */

int 
DestroyInterpCmd (ClientData clientData, 
		 Tcl_Interp * interp,
		 int argc, char *argv[])
{
  char * interp_name;
  RegisteredInterp * riPtr;

  if (argc > 2)
    {        
      sprintf(interp->result, "Too many arguments");
      return TCL_ERROR;
    }

  if (argc < 2)
    {
      interp->result = "Expected interp name but none given.";
      return TCL_ERROR;
    }

  interp_name = argv[1];
  riPtr = Soar_GetRegisteredInterp(interp_name);
  
  if (riPtr == NULL) 
    {
      sprintf(interp->result, 
	      "Attempt to delete non-existant interpreter (%s) ignored.", 
	      interp_name);
      return TCL_ERROR;      	      
    }

  if (riPtr->interp == interp)
    {
      sprintf(interp->result, 
	      "Attempt to delete current interpreter (%s) ignored.", 
	      interp_name);
      return TCL_ERROR;      	      
    }
  
  if (riPtr->agentPtr != NULL)
    {
      destroy_soar_agent(riPtr->agentPtr);
      sprintf(interp->result, "Destroying agent %s", interp_name);
    }

  Tcl_DeleteInterp(riPtr->interp);

  Soar_DestroyRegisterEntry (riPtr);

  return TCL_OK;
}

/*
 *----------------------------------------------------------------------
 *
 * ListInterpsCmd --
 *
 *	This is the command procedure for the "list-interps" command, 
 *	which lists all interp names.
 *
 * Syntax: list-interps [type]
 *         type ::= -agent | -wish | -tclsh
 *
 * Results:
 *	Returns a standard Tcl completion code.
 *
 * Side effects:
 *      Returns a list of interp names in interp->result.
 *
 *----------------------------------------------------------------------
 */

int 
ListInterpsCmd (ClientData clientData, 
		Tcl_Interp * interp,
		int argc, char *argv[])
{
  RegisteredInterp * riPtr;

  if (argc > 2)
    {        
      sprintf(interp->result, 
	      "Too many arguments.  Usage: list-interps [type]");
      return TCL_ERROR;
    }

  for (riPtr = registry; riPtr != NULL; riPtr = riPtr->nextPtr) {
    if (   (argc == 1)
	|| (   (argc == 2)
	    && (   (   (!strcmp("-agent", argv[1]))
		    && (riPtr->agentPtr != NULL))
		|| (   (!strcmp("-wish", argv[1]))
		    && (riPtr->dispPtr != NULL)
		    && (riPtr->agentPtr == NULL))
		|| (   (!strcmp("-tclsh", argv[1]))
		    && (riPtr->dispPtr == NULL)
		    && (riPtr->agentPtr == NULL)))))
      {
	Tcl_AppendElement(interp, riPtr->name);
      }
  }

  return TCL_OK;
}

/*
 *--------------------------------------------------------------
 *
 * Tcl_SendCmd --
 *
 *	This procedure is invoked to process the "send" Tcl command.
 *	See the user documentation for details on what it does.
 *
 * Results:
 *	A standard Tcl result.
 *
 * Side effects:
 *	See the user documentation.
 *
 *--------------------------------------------------------------
 */

int
Tcl_SendCmd(clientData, interp, argc, argv)
    ClientData clientData;		/* Information about sender (only
					 * dispPtr field is used). */
    Tcl_Interp *interp;			/* Current interpreter. */
    int argc;				/* Number of arguments. */
    char **argv;			/* Argument strings. */
{
    RegisteredInterp *riPtr;
    char *destName;
    int result, c, async, i, firstArg;
    size_t length;
    Tcl_DString request;

    /*
     * Since this function is invoked only when Tk is not included
     * in the interpreter, we can assume that interpreter doesn't
     * have any associated X window data for a send operation.
     */

    /*
     * Process options, if any.
     */

    async = 0;

    for (i = 1; i < (argc-1); ) {
	if (argv[i][0] != '-') {
	    break;
	}
	c = argv[i][1];
	length = strlen(argv[i]);
	if ((c == 'a') && (strncmp(argv[i], "-async", length) == 0)) {
	    async = 1;
	    i++;
	} else if ((c == 'd') && (strncmp(argv[i], "-displayof",
		length) == 0)) {
	  /* Ignore this option since it assumes an existing window */
	  i += 2;
	} else if (strcmp(argv[i], "--") == 0) {
	    i++;
	    break;
	} else {
	    Tcl_AppendResult(interp, "bad option \"", argv[i],
		    "\": must be -async, -displayof, or --", (char *) NULL);
	    return TCL_ERROR;
	}
    }

    if (argc < (i+2)) {
	Tcl_AppendResult(interp, "wrong # args: should be \"", argv[0],
		" ?options? interpName arg ?arg ...?\"", (char *) NULL);
	return TCL_ERROR;
    }

    destName = argv[i];
    firstArg = i+1;

    /*
     * See if the target interpreter is local.  If so, execute
     * the command directly.
     * The only tricky thing is passing the result from the target
     * interpreter to the invoking interpreter.  Watch out:  they
     * could be the same!
     */

    for (riPtr = registry; riPtr != NULL; riPtr = riPtr->nextPtr) {
	if (strcmp(riPtr->name, destName) != 0) {
	    continue;
	}
#ifdef USE_TK
	Tk_Preserve((ClientData) riPtr);
#endif
	if (firstArg == (argc-1)) {
	    result = Tcl_GlobalEval(riPtr->interp, argv[firstArg]);
	} else {
	    Tcl_DStringInit(&request);
	    Tcl_DStringAppend(&request, argv[firstArg], -1);
	    for (i = firstArg+1; i < argc; i++) {
		Tcl_DStringAppend(&request, " ", 1);
		Tcl_DStringAppend(&request, argv[i], -1);
	    }
	    result = Tcl_GlobalEval(riPtr->interp, Tcl_DStringValue(&request));
	    Tcl_DStringFree(&request);
	}
	if ((riPtr->interp != NULL) && (interp != riPtr->interp)) {
	    if (result == TCL_ERROR) {
		/*
		 * An error occurred, so transfer error information from the
		 * destination interpreter back to our interpreter.  Must clear
		 * interp's result before calling Tcl_AddErrorInfo, since
		 * Tcl_AddErrorInfo will store the interp's result in errorInfo
		 * before appending riPtr's $errorInfo;  we've already got
		 * everything we need in riPtr's $errorInfo.
		 */

		Tcl_ResetResult(interp);
		Tcl_AddErrorInfo(interp, Tcl_GetVar2(riPtr->interp,
			"errorInfo", (char *) NULL, TCL_GLOBAL_ONLY));
		Tcl_SetVar2(interp, "errorCode", (char *) NULL,
			Tcl_GetVar2(riPtr->interp, "errorCode", (char *) NULL,
			TCL_GLOBAL_ONLY), TCL_GLOBAL_ONLY);
	    }
	    if (riPtr->interp->freeProc != NULL) {
		interp->result = riPtr->interp->result;
		interp->freeProc = riPtr->interp->freeProc;
		riPtr->interp->freeProc = 0;
	    } else {
		Tcl_SetResult(interp, riPtr->interp->result, TCL_VOLATILE);
	    }
	    Tcl_ResetResult(riPtr->interp);
	}
#ifdef USE_TK
	Tk_Release((ClientData) riPtr);
#endif
	return result;
    }

    /*
     * If the destination wasn't local and processed above, then we 
     * have an error.
     */

    riPtr = Soar_GetRegisteredInterpByInterp (interp);
    sprintf(interp->result,
	    "send from %s: Cannot find interp named '%s'.",
	    riPtr->name, destName);
    return TCL_ERROR;
  }

/*
 *----------------------------------------------------------------------
 *
 * ScheduleInterpCmd --
 *
 *	This is the command procedure for the "schedule-interp" command, 
 *	which selects a Tcl interpreter to be scheduled with the Soar
 *      agents in the default Soar scheduler.
 *
 * Syntax:  schedule-interp {-add|-remove} {-before|-after} 
 *                          [interp_name script]
 *
 * Results:
 *	Returns a standard Tcl completion code.
 *
 * Side effects:
 *	Selects/Removes a Tcl interpreter to run in the Soar scheduler.
 *
 *----------------------------------------------------------------------
 */

int ScheduleInterpCmd (ClientData clientData, 
		       Tcl_Interp * interp,
		       int argc, char *argv[])
{
  if (argc > 5)
    {
      interp->result = "Too many arguments";
      return TCL_ERROR;
    }

  if (argc < 3)
    {
      interp->result = "Too few arguments";
      return TCL_ERROR;
    }

  if (string_match("-add", argv[1]))
      {
	RegisteredInterp * riPtr;

	if (argc < 5)
	  {
	    interp->result = "Too few arguments";
	    return TCL_ERROR;
	  }

	riPtr = Soar_GetRegisteredInterp(argv[3]);
	if (riPtr == NULL)
	  {
	    sprintf(interp->result,
		    "Unknown interpreter name given to %s %s %s: %s",
		    argv[0], argv[1], argv[2], argv[3]);
	    return TCL_ERROR;
	  }	  

	if (string_match("-before", argv[2]))
	  {

	    Soar_BeforeSchedule.interp = riPtr->interp;
	    Soar_BeforeSchedule.script = savestring(argv[4]);
	  }
	else if (string_match("-after", argv[2]))
	  {
	    Soar_AfterSchedule.interp = riPtr->interp;
	    Soar_AfterSchedule.script = savestring(argv[4]);
	  }
	else
	  {
	    sprintf(interp->result,
		    "Unrecognized argument to %s %s: %s", 
		    argv[0], argv[1], argv[2]);
	    return TCL_ERROR;
	  }
      }
  else if (string_match("-remove", argv[1]))
      {
	if (string_match("-before", argv[2]))
	  {
	    Soar_BeforeSchedule.interp = NULL;
	    if (Soar_BeforeSchedule.script != NULL)
	      {
		free(Soar_BeforeSchedule.script);
		Soar_BeforeSchedule.script = NULL;
	      }
	  }
	else if (string_match("-after", argv[2]))
	  {
	    Soar_AfterSchedule.interp = NULL;
	    if (Soar_AfterSchedule.script != NULL)
	      {
		free(Soar_AfterSchedule.script);
		Soar_AfterSchedule.script = NULL;
	      }
	  }
	else
	  {
	    sprintf(interp->result,
		    "Unrecognized argument to %s %s: %s", 
		    argv[0], argv[1], argv[2]);
	    return TCL_ERROR;
	  }
      }
  else
      {
	sprintf(interp->result,
		"Unrecognized argument to %s: %s", argv[0], argv[1]);
	return TCL_ERROR;
      }

  return TCL_OK;
}

/*
 *----------------------------------------------------------------------
 *
 * SoarnewsCmd --
 *
 *      This is the command procedure for the "soarnews" command, 
 *      which prints out information about the current Soar release.
 *
 * Syntax:  soarnews
 *
 * Results:
 *      Returns a standard Tcl completion code.
 *
 * Side effects:
 *      Prints the latest Soar news.
 *
 *----------------------------------------------------------------------
 */

int SoarnewsCmd (ClientData clientData, 
		 Tcl_Interp * interp,
		 int argc, char *argv[])
{
  /* BUGBUG update soarnews printout on successive versions */

  Tcl_AppendResult(interp,
		   "News for Soar version ", soar_version_string, "\n", 
		   soar_news_string, (char *) NULL);

  return TCL_OK;
}

/*
 *----------------------------------------------------------------------
 *
 * VersionCmd --
 *
 *      This is the command procedure for the "version" command, 
 *      which prints information about the current Soar version
 *      being run.
 *
 * Syntax:  version
 *
 * Results:
 *      Returns a standard Tcl completion code.
 *
 * Side effects:
 *      Returns the version number in the interpreter result.
 *
 *----------------------------------------------------------------------
 */
int VersionCmd (ClientData clientData, 
		Tcl_Interp * interp,
		int argc, char *argv[])
{
  if (argc > 1)
    {
      interp->result = "Too many arguments, should be: version";
      return TCL_ERROR;
    }

  sprintf(interp->result, "%s", soar_version_string); 
  return TCL_OK;
}
