/* ----------------------------------------------------------------------
 * CFOR-IO-UTILS.C
 *
 * Author: Scott Huffman
 * Date: 7/9/92
 * Purpose: A group of utility functions and data structures to make IO
 *          with Soar6b easier.  Inspired by Soarsim.
 *
 * Author: Dave Keirsey
 * Date: 12/18/92
 * Modification: Modified for ModSAF/IFOR
 *
 * Author: Karl Schwamb
 * Date: 12 May 1993
 * Modification: Enhanced utils by adding capability to assert arbitrary
 *               tree structure vice previous flat lists.  This is
 *               handled by the addition of the functions setAVobj
 *               which creates a new obj to store as the value of the
 *               specified attribute, and setAVpath_int which allows the
 *               value at the end of an attribute path to be updated.
 *
 *               Moved make_sim_obj routine from safsoar.c to here since
 *               it was misplaced in that file.
 *
 *
 * Author: Frank Koss
 * Date: 4 November 1993
 * Modification: Added removeAV() which allows for the removing of an
 *               attribute (and its value) from an obj on the input link.
 *
 * Author: Frank Koss
 * Date: 6 November 1993
 * Modification: Added setAVmult* which allow for multiple attributes
 *               with the same name.
 *
 * Author: Randy Hill
 * Date: 5 October 1995
*------------------------------------------------------------------------ */


#include "soar.h"
#include <stdtypes.h>

#include "libcforsoar_local.h"

#ifdef USE_STDARGS
#include <stdarg.h>
#else
#include <varargs.h>
#endif

/* -----------------------------------------------------------------------

   Creation functions.

- ------------------------------------------------------------------------ */

sim_object * make_sim_obj (void)
{
  sim_object * sim_obj;

  sim_obj = (sim_object *) malloc (sizeof(sim_object));

  sim_obj->AVlist             = NULL;
/*  sim_obj->needs_internalized = TRUE; */

  return sim_obj;
}


simAV * AV_alloc(void)
{
  return((simAV *) malloc(sizeof(simAV)));
}


/* -----------------------------------------------------------------------
                             Global Variables

    Pointers to certain identifiers & wmes in the current input structure
       that we need to update the input structure later.
- ------------------------------------------------------------------------ */

wme *top_input_link_wme;    /* wme: (S1 ^input-link I37) */


/*----------------------------------------------------------------------*/
/* add_sim_object_to_input_link:
   adds the sim-object to the input link (no attrs). */


void add_sim_object_to_input_link (Symbol *top_input_link_id,
                                   char * attr_name,
                                   sim_object *obj)
{
  Symbol *attr_sym;

  attr_sym = get_io_sym_constant (attr_name);
  obj->object_id = get_new_io_identifier('Z');
  obj->wme_referring_to_object
    = add_input_wme(top_input_link_id, attr_sym, obj->object_id);
  release_io_symbol(attr_sym);
}


/* GetAV gets the value of an attr for an object. */

Symbol *getAV(sim_object obj, Symbol *attr) {
  simAV *curr;

  curr = obj.AVlist;
  while (curr != NULL) {
    if (curr->attr == attr) return(curr->value);
    else curr = curr->next;
  }
  return(NIL);
}


/* GetAVstring gets the value of an attr (given as a string) for an object. */

Symbol *getAVstring(sim_object obj, char *attr) {
  simAV *curr;
  Symbol *attr_sym;

  attr_sym = get_io_sym_constant(attr);
  /* Do I need to release this symbol?  OR will that cause other ref's to it
     to vanish? */

  curr = obj.AVlist;
  while (curr != NULL) {
    if (curr->attr == attr_sym) return(curr->value);
    else curr = curr->next;
  }
  
  release_io_symbol(attr_sym);
  return(NIL);
}


/* GetAVentry returns the whole AV-entry for the requested attr. */

simAV *getAVentry(sim_object *obj, Symbol *attr) {
  simAV *curr;

  curr = obj->AVlist;
  while (curr != NULL) {
    if (curr->attr == attr) return(curr);
    else curr = curr->next;
  }
  return(NULL);
}

/* SetAV sets the value of an attr for an object.
   If the object already has the attr, its value is updated.  If not,
   a new attr is created.
 */

simAV *
setAV(sim_object *obj, Symbol *attr, Symbol *val) {
  simAV *temp;

  temp = getAVentry(obj,attr);

  if (temp != NULL) {

        release_io_symbol(attr);

        if (val != temp->value) {        /* There is a previous value. */

            release_io_symbol(temp->value);
            temp->value = val;
            temp->changed = TRUE;
        }
        else
          {
            release_io_symbol(val);
          }
    }

    /* There is no previous value. */
    else {

        temp = AV_alloc();
        temp->next = obj->AVlist;
        obj->AVlist = temp;
        temp->attr = attr;
        temp->value = val;
        temp->has_wme = FALSE;
        temp->changed = TRUE;
        temp->public = TRUE;
        temp->object = NULL;
    }

    return temp;
}
 
/* SetAVmult sets the value of an attr for an object.  In all cases, a
   new attribute is created, even if the attribute already exists.
 */

simAV *
setAVmult(sim_object *obj, Symbol *attr, Symbol *val)
{
  simAV *temp;

  /* Make a new attribute */
  temp = AV_alloc();
  temp->next = obj->AVlist;
  obj->AVlist = temp;
  temp->attr = attr;
  temp->value = val;
  temp->has_wme = FALSE;
  temp->changed = TRUE;
  temp->public = TRUE;
  temp->object = NULL;

  return temp;
}

/* setAVmult_string allows you to add a new attr-value using strings. */
void setAVmult_string(sim_object *obj, char *attr, char *val) {
  Symbol *attr_sym, *val_sym;
  
  attr_sym = get_io_sym_constant(attr);
  val_sym = get_io_sym_constant(val);
  setAVmult(obj,attr_sym,val_sym);
}

/* setAVmult_obj allows you to add a new attr-object pair using a
   string for the attr.  A new attribute is created rather than
   replacing an existing one.
 */
sim_object *
setAVmult_obj (sim_object * obj, char * attr_name)
{
  Symbol * attr_sym;
  Symbol * object_sym;
  simAV * av_for_attr;
  sim_object * so_for_obj;

  attr_sym   = get_io_sym_constant(attr_name);
  object_sym = get_new_io_identifier('Y');

  av_for_attr = setAVmult(obj, attr_sym, object_sym);
  so_for_obj  = make_sim_obj();

  so_for_obj->object_id = object_sym;
  av_for_attr->object = so_for_obj;

  return so_for_obj;
}

/* setAVstring allows you to add a new attr-value using strings. */
void
setAVstring(sim_object *obj, char *attr, char *val) {
  Symbol *attr_sym, *val_sym;
  
  attr_sym = get_io_sym_constant(attr);
  val_sym = get_io_sym_constant(val);
  setAV(obj,attr_sym,val_sym);
}

 
/* setAVint allows you to add a new attr-value using a string for the attr
   and an INT for the value. */

void
setAVint(sim_object *obj, char *attr, long val) {
  Symbol *attr_sym, *val_sym;
  
  attr_sym = get_io_sym_constant(attr);
  val_sym = get_io_int_constant(val);
  setAV(obj,attr_sym,val_sym);
}

/* setAVfloat allows you to add a new attr-value using a string for the attr
   and a FLOAT for the value. */

void
setAVfloat(sim_object *obj, char *attr, float val) {
  Symbol *attr_sym, *val_sym;
  
  attr_sym = get_io_sym_constant(attr);
  val_sym = get_io_float_constant(val);
  setAV(obj,attr_sym,val_sym);
}

/* setAVobj alloes you to add a new attr-object pair using a string for the
   attr
*/

sim_object *
setAVobj (sim_object * obj, char * attr_name)
{
  Symbol * attr_sym;
  Symbol * object_sym;
  simAV * av_for_attr;
  sim_object * so_for_obj;

  attr_sym   = get_io_sym_constant(attr_name);
  object_sym = get_new_io_identifier('X');

#ifdef DEBUG_SYMBOLS
  print("SetAVobj: ^%s X%d\n", attr_name, object_sym->id.name_number);
#endif

  av_for_attr = setAV(obj, attr_sym, object_sym);
  so_for_obj  = make_sim_obj();

  so_for_obj->object_id = object_sym;
  av_for_attr->object = so_for_obj;

  return so_for_obj;
}

/*
 * Removes the WME whose attribute name is *attribute.  Assumes *attribute's
 * value is only a value or an object (not a tree).
 */
void removeAV(sim_object *obj, char *attribute)
{
    Symbol *attr_sym = get_io_sym_constant(attribute);
    simAV  *curr = obj->AVlist,
           *prev = NULL;

    /* Go through the WMEs looking for the correct attribute */
    while (curr != NULL) {

        /* If found correct attribute, remove it */
        if (curr->attr == attr_sym) {

            if (prev == NULL)                /* First in list */
                obj->AVlist = curr->next;
            else                             /* Not first in list */
                prev->next = curr->next;
            if (curr->has_wme)
                remove_input_wme(curr->wme);
            if (curr->object != NULL) {
                release_io_symbols_simobj(curr->object);
                free(curr->object);
            }
            else {
                release_io_symbol(curr->value);
            }
            release_io_symbol(curr->attr);

            free(curr);
            release_io_symbol(attr_sym);
            return;
        }
        prev = curr;
        curr = curr->next;
    }
    release_io_symbol(attr_sym);
}

void removeAVobjTree(sim_object *obj, char *attribute)
{
  Symbol *attr_sym = get_io_sym_constant(attribute);
  simAV  *curr = obj->AVlist,
         *prev = NULL;

  /* Go through the WMEs looking for the correct attribute */
  while (curr != NULL) {

    /* If found correct attribute, remove object */
    if (curr->attr == attr_sym)
      {
        if (prev == NULL)                /* First in list */
          obj->AVlist = curr->next;
        else                             /* Not first in list */
          prev->next = curr->next;
        if (curr->has_wme)
          remove_input_wme(curr->wme);
        if (curr->object != NULL) {
          release_io_symbols_simtree(curr->object);
        } else {
          release_io_symbol(curr->value);
        }
        release_io_symbol(curr->attr);
        
        free(curr);
        release_io_symbol(attr_sym);
        return;
      }
    prev = curr;
    curr = curr->next;
  }
  
  release_io_symbol(attr_sym);
}

/*
 * REMOVEAV_OBJ
 *
 * Releases and frees all symbols and memory associated with *obj.
 */
void removeAV_obj(sim_object *obj)
{
    simAV *curr,
          *temp;

    if (obj == NULL)
        return;

    /* Release the symbol for this object */
    release_io_symbol(obj->object_id);

    /* Free the AV list recursively removing objects hung off of it */
    curr = obj->AVlist;
    while (curr != NULL) {
        release_io_symbol(curr->attr);
        if (curr->has_wme)
            remove_input_wme(curr->wme);
        if (curr->object != NULL)
            removeAV_obj(curr->object);
        else
            release_io_symbol(curr->value);

        temp = curr->next;
        free(curr);
        curr = temp;
    }
    free(obj);
}


/* This function is used to set a value at the end of an attr path. */
/* For example, in Soar a clause might be:
     (<s> ^foo.bar.baz 42)
   To access and set the same attribute the following call would
   suffice:
     setAVpath_int(obj, "foo", "bar", "baz", NULL, 42);
*/

#ifdef USE_STDARGS

/*VARARGS1*/

void
setAVpath_int (sim_object * obj, ...) {
    va_list args;
    char * attr_name;
    char * next_attr_name;
    Symbol * attr_sym;
    simAV * av_for_attr;
    int value;

    va_start (args, obj);

#else

/*VARARGS*/

void
setAVpath_int (va_alist) va_dcl {
    va_list args;
    sim_object * obj;
    char * attr_name;
    char * next_attr_name;
    Symbol * attr_sym;
    simAV * av_for_attr;
    int value;

    va_start (args);
    obj = va_arg(args, sim_object *);

#endif /* USE_STDARGS */
    
    attr_name = va_arg(args, char *);

    while (next_attr_name = va_arg(args, char *))
      {
        attr_sym = get_io_sym_constant(attr_name);
        av_for_attr = getAVentry(obj, attr_sym);
        release_io_symbol(attr_sym);
        obj = av_for_attr->object;
        attr_name = next_attr_name;
      }
    value = va_arg(args, int);

    setAVint(obj, attr_name, value);

    va_end (args);
  }


/* This function is used to set a value at the end of an attr path. */
/* For example, in Soar a clause might be:
     (<s> ^foo.bar.baz 42.0)
   To access and set the same attribute the following call would
   suffice:
     setAVpath_float(obj, "foo", "bar", "baz", NULL, 42.0);
*/
#ifdef USE_STDARGS

/*VARARGS1*/

void
setAVpath_float (sim_object * obj, ...) {
    va_list args;
    char * attr_name;
    char * next_attr_name;
    Symbol * attr_sym;
    simAV * av_for_attr;
    float value;

    va_start (args, obj);

#else

/*VARARGS*/

void
setAVpath_float (va_alist) va_dcl {
    va_list args;
    sim_object * obj;
    char * attr_name;
    char * next_attr_name;
    Symbol * attr_sym;
    simAV * av_for_attr;
    float value;

    va_start (args);
    obj = va_arg(args, sim_object *);

#endif /* USE_STDARGS */
    
    attr_name = va_arg(args, char *);

    while (next_attr_name = va_arg(args, char *))
      {
        attr_sym = get_io_sym_constant(attr_name);
        av_for_attr = getAVentry(obj, attr_sym);
        release_io_symbol(attr_sym);
        obj = av_for_attr->object;
        attr_name = next_attr_name;
      }
    value = (float) va_arg(args, double);

    setAVfloat(obj, attr_name, value);

    va_end (args);
  }

/* This function is used to set a string value at the end of an attr path. */
/* For example, in Soar a clause might be:
     (<s> ^foo.bar.baz hello)
   To access and set the same attribute the following call would
   suffice:
     setAVpath_string(obj, "foo", "bar", "baz", NULL, "hello");
*/

#ifdef USE_STDARGS

/*VARARGS1*/

  void
  setAVpath_string (sim_object * obj, ...) {
    va_list args;
    char * attr_name;
    char * next_attr_name;
    Symbol * attr_sym;
    simAV * av_for_attr;
    char *value;

    va_start (args, obj);

#else

/*VARARGS*/

  void
  setAVpath_string (va_alist) va_dcl {
    va_list args;
    sim_object * obj;
    char * attr_name;
    char * next_attr_name;
    Symbol * attr_sym;
    simAV * av_for_attr;
    char *value;

    va_start (args);
    obj = va_arg(args, sim_object *);

#endif /* USE_STDARGS */
    
    attr_name = va_arg(args, char *);

    while (next_attr_name = va_arg(args, char *))
      {
        attr_sym = get_io_sym_constant(attr_name);
        av_for_attr = getAVentry(obj, attr_sym);
        release_io_symbol(attr_sym);
        obj = av_for_attr->object;
        attr_name = next_attr_name;
      }
    value = va_arg(args, char *);

    setAVstring(obj, attr_name, value);

    va_end (args);
  }

  
/* ----------------------------------------------------------------------
   Internalize_world_object
   ---------------------------------------------------------------------- */

void mark_object_as_not_internalized (sim_object *obj) {
  simAV *curr;

  curr = obj->AVlist;

  while (curr != NULL) {
    curr->changed = TRUE;
    curr->has_wme = FALSE;

    if (curr->object)
      mark_object_as_not_internalized(curr->object);

    curr = curr->next;
  }
}


void internalize_world_object(sim_object *obj) {
  simAV *curr;

/*  if (!obj->needs_internalized) return; */

  curr = obj->AVlist;

  while (curr != NULL) {
    if ((curr->changed == TRUE) && (curr->public == TRUE)) {
      if (curr->has_wme) remove_input_wme(curr->wme);
      curr->wme = add_input_wme(obj->object_id,
                                curr->attr,
                                curr->value);
      curr->changed = FALSE;
      curr->has_wme = TRUE;
    }

    if (curr->object)
      internalize_world_object(curr->object);

    curr = curr->next;
  }

/*  obj->needs_internalized = FALSE; */
}
    

/* ----------------------------------------------------------------------
   Make_private_AV(obj,attrstring):

     declares attr (w/string name attrstring) of obj a private attr.

   ---------------------------------------------------------------------- */

void make_private_AV(sim_object *obj, char *attr) {
  simAV *temp;
  Symbol *attr_sym;

  attr_sym = get_io_sym_constant(attr);

  temp = getAVentry(obj,attr_sym);
  if (temp != NULL) {
    /* There is a previous value. */
    temp->public = FALSE;
  }

  /* relase attr_sym??? But it's stored away... */
}


/* ----------------------------------------------------------------------
   RELEASE_IO_SYMBOLS_SIMOBJ(simobj)
 
   Releases all IO symbols assoc. w/simobj.

   ---------------------------------------------------------------------- */

void release_io_symbols_simobj(sim_object * obj) {
  simAV *curr;

  release_io_symbol(obj->object_id);

  curr = obj->AVlist;

  while (curr != NULL) {
    release_io_symbol(curr->attr);
    release_io_symbol(curr->value);
    curr = curr->next;
  }
}

void release_io_symbols_simtree(sim_object *obj) {
  simAV *curr, *next;

  curr = obj->AVlist;

  while (curr != NULL)
    {
      if (curr->object)
        release_io_symbols_simtree(curr->object);    /* Recurse */
      else
        release_io_symbol(curr->value);

      next = curr->next;

      release_io_symbol(curr->attr);

      free(curr);

      curr = next;
    }

  release_io_symbol(obj->object_id);
  free(obj);
}

 
/* NOTE: Multiply defined in database.c too... */
/*

Symbol *get_output_value (io_wme *outputs, Symbol *id, Symbol *attr) {
  io_wme *iw;

  for (iw=outputs; iw!=NIL; iw=iw->next)
    if ( ((id==NIL)||(id==iw->id)) &&
         ((attr==NIL)||(attr==iw->attr)) ) return iw->value;
  return NIL;
}

*/

